#include "config.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <stdio.h>
#include <pwd.h>
#include <errno.h>
#include <string.h>

#if HAVE_GETOPT_H
#include <getopt.h>
#endif
#if HAVE_UNISTD_H
#include <unistd.h>
#endif




#include <syslog.h>
#include <signal.h>

#include "misc.h"
#include "get_uid.h"

#define FATAL "hidentd: fatal:"
#define INFO  "hidentd: info:"

struct ip_address la, ra;
uint32 lp, rp;
char flagverbose=0, flagsyslog=1, flagprivate=0;

void fail(char *msg) {
	if(flagsyslog) 
 	 syslog(LOG_DAEMON|LOG_ERR,"%s: %s\n",msg,strerror(errno));
	else
	 fprintf(stderr,"%s %s: %s\n",FATAL,msg,strerror(errno));
	exit(1);
}

void log(struct conn_info c, char *m1, char *m2) {
char local[20], remote[20];

	sprintf(local,"%d.%d.%d.%d",c.laddr.d[0],c.laddr.d[1],c.laddr.d[2],c.laddr.d[3]);
	sprintf(remote,"%d.%d.%d.%d",c.raddr.d[0],c.raddr.d[1],c.raddr.d[2],c.raddr.d[3]);
	if(flagsyslog)
		syslog(LOG_DAEMON|LOG_INFO,"query: %s:%d <-> %s:%d, %s %s",local,c.lport,remote,c.rport,m1,m2);
	else {
		fprintf(stderr,"%s query: %s:%d <-> %s:%d, %s %s\n",INFO,local,c.lport,remote,c.rport,m1,m2);
		fflush(stderr);
	}
}
		




void ident_reply(struct conn_info c, char *reply) {
	printf("%d, %d : USERID : UNIX :%s\r\n",c.lport,c.rport,reply);
}

void ident_error(struct conn_info c, char *error) {
	printf("%d, %d : ERROR : %s\r\n",c.lport,c.rport,error);
}


void get_addrs(struct conn_info *c) {
struct sockaddr_in local, remote;
int len;

	len=sizeof(local);
	if(getsockname(0,(struct sockaddr *)&local, &len)==-1) {
		fail("could not get local address");
		exit(1);
	}
	memcpy(c->laddr.d,&local.sin_addr,sizeof(c->laddr.d));
	len=sizeof(remote);
	if(getpeername(0,(struct sockaddr *)&remote, &len)==-1) {
		fail("could not get remote address");
		exit(1);
	}
	memcpy(c->raddr.d,&remote.sin_addr,sizeof(c->raddr.d));
}

void get_ports(struct conn_info *c) {
char *x, line[512], *lport, *rport;
int len, fields[2];
uint32 l, r;

	x=fgets(line,sizeof(line),stdin);
	if(!x) {
		log(*c,"dropping: could not read from network: ",strerror(errno));
		exit(1);
	}
	if(line[strlen(line)-1]!='\n') {
		log(*c,"dropping: input too long","");
		exit(1);
	}
	len=strlen(line);
	tokenize(line,fields,2);
	if((fields[1]>=len)) {
		log(*c,"dropping: input too short","");
		exit(1);
	}
	lport=&line[fields[0]]; rport=&line[fields[1]];
	if((lport[scan_uint32(lport,&l)])||(l>65535)) {
		ident_error(*c,"INVALID-PORT");
		log(*c,"malformed input: could not parse local-port",""); 
		exit(1);
	}
	c->lport=l;
	if((rport[scan_uint32(rport,&r)])||(r>65535)) {
		ident_error(*c,"INVALID-PORT");
		log(*c,"malformed input: could not parse remote-port","");
		exit(1);
	}
	c->rport=r;
}

void timeout(int s) {
	fail("dropping connection: timeout");
	exit(1);
}

void do_help() {
	fprintf(stderr,"Usage: hidentd [-hsv] [-u username] [-p username]\n");
	fprintf(stderr,"\n\t-h\t\thelp\n");
	fprintf(stderr,"\t-v\t\tverbose\n");
	fprintf(stderr,"\t-s\t\tlog to stderr\n");
	fprintf(stderr,"\t-u username\tanswer with 'username' for masqueraded connections\n");
	fprintf(stderr,"\t-p username\talways answer with 'username' (overrides -u)\n");
	fprintf(stderr,"\nSend bug reports to %s\n", PACKAGE_BUGREPORT);
	exit(0);
}

int main(int argc, char **argv) {
	char *username=NULL, *allname=NULL;
	int uid=-1, opt;
	struct passwd *pw;
	struct conn_info ci;

	while((opt=getopt(argc, argv, "hvp:u:s"))!=-1) {
	 switch(opt) {
		 case 'h': do_help(); break;
		 case 'v': flagverbose=1; break;
		 case 'u': username=optarg; break;
		 case 's': flagsyslog=0;break;
		 case 'p': allname=optarg; break;
		 default:  do_help();
	 }
	}
	memset(&ci,0,sizeof(ci));
	get_addrs(&ci);
	alarm(60);
	signal(SIGALRM,timeout);
	get_ports(&ci);
	if(allname) 
		username=allname;
	else {
		uid=get_uid(ci);
		if((uid<0)&&(!username)) {
				log(ci,"error: could not read /proc/tcp:",strerror(errno));
				ident_error(ci,"NO-USER");
				exit(1);
		} else if(uid>=0) {
			pw=getpwuid(uid);
			if(!pw) {
				log(ci,"error: could not get username:",strerror(errno));
				ident_error(ci,"UNKNOWN-ERROR");
				exit(1);
			}
			username=pw->pw_name;
		}
	}
	ident_reply(ci,username);
	if(flagverbose) log(ci,"answer:",username);
	exit(0);
}
	
